import requests
from adalib_auth import config, keycloak

from .. import adaboard

ALLOWED_OCI_PROJECT_TYPES = [
    "adamatics",
    "apps",
    "base_images",
    "kernels",
    "labapps",
]
ALLOWED_METADATA_TYPES = ["app", "base", "internal", "kernel", "lab_app"]

ALLOWED_OCI_SOURCE_TYPES_OPERATIONS = {"lab": "push", "registry": "copy"}
ALLOWED_PUBLISH_LOG_SOURCES = ["publish", "system"]
HARBOR_HOST = "container-registry"
HARBOR_PORT = "80"


def archive_image(metadata_id: int) -> None:
    """
    Archive an image in the Harbor registry. This will archive the OCI image metadata object and works as a soft delete.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/archive_image.ipynb) to test this function or build upon it.

    :param metadata_id: the metadata ID of the image
    :type metadata_id: int
    :return: nothing
    :rtype: None
    """

    # Check if image is already archived
    img_metadata = adaboard.request_adaboard(
        path=f"registry/metadata/{metadata_id}",
    ).json()
    # If not archived, archive it
    if not img_metadata["archived"]:
        payload = {"archived": True}
        adaboard.request_adaboard(
            path=f"registry/metadata/{metadata_id}/archive",
            method=requests.put,
            json=payload,
        )
        return None
    else:
        raise RuntimeError(
            f"Image with metadata ID {metadata_id} is already archived."
        )


def create_image_metadata(
    source_type: str,
    source_repository: str,
    source_tag: str,
    project_name: str,
    repository_name: str,
    tag: str,
    type_id: str,
    name: str,
    description: str,
    username: str,
    cmd: str = "",
    picture_id=1,
) -> dict[str, str | int | bool | dict[str, str]]:
    """
    Create metadata for a specific image in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/create_image_metadata.ipynb) to test this function or build upon it.

    :param source_type: the location where the source image is to be pulled from
    :type source_type: str
    :param source_repository: the name of the repository of the source image
    :type source_repository: str
    :param source_tag: the tag of the source image
    :type source_tag: str
    :param project_name: the name of the target project for the image metadata
    :type project_name: str
    :param repository_name: the name of the target repository for the image metadata
    :type: repository_name: str
    :param tag: the tag for the target image metadata
    :type tag: str
    :param type_id: the type of the image
    :type type_id: str
    :param name: the name for the target image metadata
    :type name: str
    :param description: the description for the target image metadata
    :type description: str
    :param username: the current username (tip: check "LOGNAME" environment variable)
    :type username: str
    :param cmd: the startup command for the target image metadata, defaults to ""
    :type cmd: str, optional
    :param picture_id: the ID of the picture for the image, defaults to 1
    :type picture_id: int, optional

    :return: image metadata
    :rtype: dict
    """
    # TODO fix the use of environmental variables here
    adalib_config = config.get_config()

    # Make sure container options are valid
    assert (
        project_name in ALLOWED_OCI_PROJECT_TYPES
    ), f"Can only create metadata for images that are of the type {', '.join([x for x in ALLOWED_OCI_PROJECT_TYPES])}"
    assert (
        type_id in ALLOWED_METADATA_TYPES
    ), f"Can only create metadata of the type {', '.join([x for x in ALLOWED_METADATA_TYPES])}"
    assert source_type in list(
        ALLOWED_OCI_SOURCE_TYPES_OPERATIONS.keys()
    ), f"""Can only create metadata for images located in {', '.join([x for x in list(
        ALLOWED_OCI_SOURCE_TYPES_OPERATIONS.keys()
    )])}"""

    # Create metadata object in AdaLab
    payload = {
        "sha256": "sha256:00000",
        "oci_image_name": {
            "project": project_name,
            "repository": repository_name,
            "tag": tag,
            "host": f"{HARBOR_HOST}:{HARBOR_PORT}",
        },
        "type": type_id,
        "name": name,
        "description": description,
        "startup_cmd": cmd,
        "image_id": picture_id,
    }

    response_metadata = adaboard.request_adaboard(
        path="registry/metadata", method=requests.post, json=payload
    ).json()

    # Build request payload based on user and default options
    source_project = f"{project_name}_temp"

    payload = {
        "script": "pusher",
        "user_id": username,
        "pool": "internal",
        "cleanup": True,
        "timeout": 3600,
        "run_in_isolation": False,
        "start_podman_sidecar": False,
        "config": {
            "ADABOARD_API_URL": {
                "value": adalib_config.SERVICES["adaboard-api"]["external"],
                "real_value": adalib_config.SERVICES["adaboard-api"][
                    "external"
                ],
            },
            "CONTAINER_METADATA_ID": {
                "value": response_metadata["metadata_id"],
                "real_value": response_metadata["metadata_id"],
            },
            "HARBOR_HOST": {
                "value": HARBOR_HOST,
                "real_value": HARBOR_HOST,
            },
            "HARBOR_NAMESPACE": {
                "value": adalib_config.NAMESPACE,
                "real_value": adalib_config.NAMESPACE,
            },
            "HARBOR_TARGET_PROJECT": {
                "value": response_metadata["oci_image_name"]["project"],
                "real_value": response_metadata["oci_image_name"]["project"],
            },
            "HARBOR_TARGET_REPOSITORY_NAME": {
                "value": response_metadata["oci_image_name"]["repository"],
                "real_value": response_metadata["oci_image_name"][
                    "repository"
                ],
            },
            "HARBOR_TARGET_TAG": {
                "value": response_metadata["oci_image_name"]["tag"],
                "real_value": response_metadata["oci_image_name"]["tag"],
            },
            "KEEP_IMAGES": {"value": True, "real_value": True},
            "NETWORK_HOST": {
                "value": adalib_config.NETWORK_HOST,
                "real_value": adalib_config.NETWORK_HOST,
            },
            "OPERATION": {
                "value": ALLOWED_OCI_SOURCE_TYPES_OPERATIONS[source_type],
                "real_value": ALLOWED_OCI_SOURCE_TYPES_OPERATIONS[source_type],
            },
        },
    }
    if source_type.lower() == "lab":
        payload["config"]["RAW_SOURCE_IMAGE"] = {
            "value": f"{source_repository}:{source_tag}",
            "real_value": f"{source_repository}:{source_tag}",
        }
    elif source_type.lower() == "registry":
        payload["config"]["HARBOR_SOURCE_PROJECT"] = {
            "value": source_project,
            "real_value": source_project,
        }
        payload["config"]["HARBOR_SOURCE_REPOSITORY_NAME"] = {
            "value": source_repository,
            "real_value": source_repository,
        }
        payload["config"]["HARBOR_SOURCE_TAG"] = {
            "value": source_tag,
            "real_value": source_tag,
        }

    # Push image from a source to a target repo in Harbor
    response_pusher = adaboard.request_adaboard(
        path="script_runner/runs/start/", method=requests.post, json=payload
    ).json()

    # Update metadata object in AdaLab
    payload = {
        "name": response_metadata["name"],
        "description": response_metadata["description"],
        "startup_cmd": response_metadata["startup_cmd"],
        "run_id": response_pusher["id"],
    }

    adaboard.request_adaboard(
        path=f"registry/metadata/{response_metadata['metadata_id']}",
        method=requests.put,
        json=payload,
    )

    # Update image state in AdaLab to the one in Harbor
    adaboard.request_adaboard(
        path=f"registry/metadata/{response_metadata['metadata_id']}/ready",
        method=requests.put,
    ).json()

    # Get image metadata from AdaLab
    response = adaboard.request_adaboard(
        path=f"registry/metadata/{response_metadata['metadata_id']}",
    ).json()

    return response


def edit_image_metadata(
    metadata_id: int,
    new_name: str = "",
    new_description: str = "",
    new_cmd: str = "",
    new_run_id: int = 0,
    new_picture_id: int = 1,
) -> None:
    """
    Edit the metadata of an image in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/edit_image_metadata.ipynb) to test this function or build upon it.

    :param metadata_id: the metadata ID of the image
    :type metadata_id: int
    :param new_name: the new name of the image, defaults to ""
    :type image_name: str, optional
    :param new_description: the new description of the image, defaults to ""
    :type new_description: str, optional
    :param new_cmd: the new startup command of the image, defaults to ""
    :type new_cmd: str, optional
    :param new_run_id: the new run ID of the image, defaults to 0
    :type new_run_id: int, optional
    :param new_picture_id: the ID of the new picture of the image, defaults to 1
    :type new_picture_id: int, optional
    :return: nothing
    :rtype: None
    """

    # Build request payload based on new and old options
    old_metadata = adaboard.request_adaboard(
        path=f"registry/metadata/{metadata_id}",
    ).json()

    if not new_name:
        new_name = old_metadata["name"]
    if not new_description:
        new_description = old_metadata["description"]
    if not new_cmd:
        new_cmd = old_metadata["startup_cmd"]
    if not new_run_id:
        new_run_id = old_metadata["run_id"]
    if not new_picture_id:
        new_picture_id = old_metadata["image_id"]

    payload = {
        "name": new_name,
        "description": new_description,
        "startup_cmd": new_cmd,
        "run_id": new_run_id,
        "image_id": new_picture_id,
    }

    # Edit metadata
    adaboard.request_adaboard(
        path=f"registry/metadata/{metadata_id}",
        method=requests.put,
        json=payload,
    )

    return None


def get_all_metadata(
    include_archived: bool = False,
) -> list[dict[str, str | int | bool | dict[str, str]]]:
    """
    Get metadata for all images in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_all_metadata.ipynb) to test this function or build upon it.

    :param include_archived: whether to include archived images, defaults to False
    :type include_archived: bool, optional
    :return: metadata for all images
    :rtype: list
    """
    response = adaboard.request_adaboard(
        path="registry/metadata", params={"get_archived": False}
    ).json()
    if include_archived:
        res_archived = adaboard.request_adaboard(
            path="registry/metadata", params={"get_archived": True}
        ).json()
        response.extend(res_archived)

    return response


def get_image_digest(metadata_id: str) -> str:
    """
    Get the digest of an image in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_image_digest.ipynb) to test this function or build upon it.

    :param metadata_id: the metadata ID of the image
    :type metadata_id: int
    :return: image digest for the remote image
    :rtype: str
    """

    response = adaboard.request_adaboard(
        path=f"registry/metadata/{metadata_id}",
    ).json()

    return response["sha256"]


def get_image_id(project_name: str, repository_name: str, tag: str) -> int:
    """
    Get the ID of an image in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_image_id.ipynb) to test this function or build upon it.

    :param project_name: the name of the project the image belongs to
    :type project_name: str
    :param repository_name: the name of the image repository
    :type: repository_name: str
    :param tag: the tag of the specific image
    :type tag: str
    :return: image ID
    :rtype: int
    """
    # Make sure image options are valid
    assert (
        project_name in ALLOWED_OCI_PROJECT_TYPES
    ), f"Can only list IDs for images that are of the type {', '.join([x for x in ALLOWED_OCI_PROJECT_TYPES])}"

    # Get all images from the repository, page by page
    all_images = adaboard.get_all_pages(
        api_url=f"registry/{project_name}/{repository_name}/images/v2/"
    )
    # Find the specific image and return its ID
    for image in all_images:
        if image["tags"][0] == tag:
            return image["id"]

    raise ValueError(
        f"Could not find image with tag {tag} in repository {repository_name}"
    )


def get_image_metadata(
    metadata_id: int,
) -> dict[str, str | int | bool | dict[str, str]]:
    """
    Get metadata for a specific image in the Harbor registry based on its ID.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_image_metadata.ipynb) to test this function or build upon it.

    :param metadata_id: the metadata ID of the image
    :type metadata_id: int
    :return: image metadata
    :rtype: dict
    """

    response = adaboard.request_adaboard(
        path=f"registry/metadata/{metadata_id}",
    ).json()

    return response


def get_image_metadata_id(
    project_name: str, repository_name: str, tag: str
) -> int:
    """
    Get the metadata ID for a specific image in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_image_metadata_id.ipynb) to test this function or build upon it.

    :param project_name: the name of the project the image belongs to
    :type project_name: str
    :param repository_name: the name of the image repository
    :type: repository_name: str
    :param tag: the tag of the specific image
    :type tag: str
    :return: image metadata ID
    :rtype: int
    """

    # Make sure image options are valid
    assert (
        project_name in ALLOWED_OCI_PROJECT_TYPES
    ), f"Can only list metadata IDs for images that are of the type {', '.join([x for x in ALLOWED_OCI_PROJECT_TYPES])}"

    # Get metadata for all images
    response = adaboard.request_adaboard(
        path="registry/metadata", params={"get_archived": False}
    ).json()
    res_archived = adaboard.request_adaboard(
        path="registry/metadata", params={"get_archived": True}
    ).json()
    response.extend(res_archived)

    # Find the specific image and return its metadata
    for image in response:
        if (
            image["oci_image_name"]["project"] == project_name
            and image["oci_image_name"]["repository"] == repository_name
            and image["oci_image_name"]["tag"] == tag
        ):
            return image["metadata_id"]

    raise ValueError(
        f"Could not find image with tag {tag} in repository {repository_name}"
    )


def get_project_repositories(project_name: str) -> list[str]:
    """
    Get all image repositories under a given project name.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_project_repositories.ipynb) to test this function or build upon it.

    :param project_name: the name of the project to list images from
    :type project_name: str
    :return: list of existing images under a given project name
    :rtype: list
    """

    # Make sure project options are valid
    assert (
        project_name in ALLOWED_OCI_PROJECT_TYPES
    ), f"Can only list images that are of the type {', '.join([x for x in ALLOWED_OCI_PROJECT_TYPES])}"

    # Get all repositories from the project, page by page
    all_images = adaboard.get_all_pages(
        api_url=f"registry/{project_name}/v2/repositories/"
    )

    return [x["name"] for x in all_images]


def get_project_stats(project_name: str) -> list[dict[str, str | int]]:
    """
    Get statistics for a specific project in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_project_stats.ipynb) to test this function or build upon it.

    :param project_name: the name of the project to get stats from
    :type project_name: str
    :return: usage statistics for a given project
    :rtype: list
    """

    # Make sure project options are valid
    assert (
        project_name in ALLOWED_OCI_PROJECT_TYPES
    ), f"Can only show stats of projects that are {', '.join([x for x in ALLOWED_OCI_PROJECT_TYPES])}"

    # Get all repositories from the project, page by page
    all_images = adaboard.get_all_pages(
        api_url=f"registry/{project_name}/v2/repositories/"
    )

    # Compile and return statistics
    return [
        {"repository": x["name"], "count": x["image_amount"]}
        for x in all_images
    ]


def get_projects(include_archived: bool = False) -> list[str]:
    """
    Get a list of the projects in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_projects.ipynb) to test this function or build upon it.

    :param include_archived: whether to include archived projects
    :type include_archived: bool
    :return: list of existing projects
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path="registry/metadata", params={"get_archived": False}
    ).json()
    if include_archived:
        res_archived = adaboard.request_adaboard(
            path="registry/metadata", params={"get_archived": True}
        ).json()
        response.extend(res_archived)

    projects = [x["oci_image_name"]["project"] for x in response]

    return list(set(projects))


def get_publish_logs(
    metadata_id: int, source: str = "publish"
) -> list[dict[str, str]]:
    """
    Get the publish or system logs of an image publishing process.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_publish_logs.ipynb) to test this function or build upon it.

    :param metadata_id: the metadata ID of the image
    :type metadata_id: int
    :param source: the source of the logs, either "build" or "publish", defaults to "publish"
    :type source: str, optional
    :return: build logs
    :rtype: list
    """

    assert (
        source in ALLOWED_PUBLISH_LOG_SOURCES
    ), f"Log source must be one of {', '.join([x for x in ALLOWED_PUBLISH_LOG_SOURCES])}"

    # Get run ID of the publishing process
    try:
        mdata = adaboard.request_adaboard(
            path="registry/metadata", params={"get_archived": False}
        ).json()
        mdata_archived = adaboard.request_adaboard(
            path="registry/metadata", params={"get_archived": True}
        ).json()
        mdata.extend(mdata_archived)
        publish_id = [
            x["run_id"] for x in mdata if x["metadata_id"] == metadata_id
        ][0]
    except IndexError:
        raise ValueError(
            f"Could not find metadata with ID {metadata_id} in the registry."
        )

    response_raw = adaboard.request_adaboard(
        path=f"script_runner/runs/{publish_id}/logs/"
    ).json()
    response = response_raw["log"]

    if source == "publish":
        logs = [
            {
                "time": res["timestamp"],
                "input": res["input"],
                "output": res["output"].replace(res["input"], ""),
            }
            for res in response
            if res["type"] == "command"
        ]
    elif source == "system":
        logs = [
            {"time": res["timestamp"], "message": res["output"]}
            for res in response
            if res["type"] == "system"
        ]

    return logs


def get_repositories_by_author(
    author_id: str,
    include_archived: bool = False,
) -> list[str]:
    """
    Get a list of the repositories in the Harbor registry by author ID.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_repositories_by_author.ipynb) to test this function or build upon it.

    :param author_id: the ID of the author to list repositories from
    :type author_id: str
    :param include_archived: whether to include archived images, defaults to False
    :type include_archived: bool, optional
    :return: list of existing repositories from a given author
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path="registry/metadata", params={"get_archived": False}
    ).json()
    if include_archived:
        res_archived = adaboard.request_adaboard(
            path="registry/metadata", params={"get_archived": True}
        ).json()
        response.extend(res_archived)

    author_repos = [x["name"] for x in response if x["user_id"] == author_id]

    return list(set(author_repos))


def get_repositories_by_type(
    type_id: str, include_archived: bool = False
) -> list[str]:
    """
    Get a list of the repositories in the Harbor registry by image type.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_repositories_by_type.ipynb) to test this function or build upon it.

    :param type: type of image
    :type type: str
    :param include_archived: whether to include archived images, defaults to False
    :type include_archived: bool, optional
    :return: list of existing repositories of a given type
    :rtype: list
    """

    # Make sure image options are valid
    assert (
        type_id in ALLOWED_METADATA_TYPES
    ), f"Can only get repositories of the type {', '.join([x for x in ALLOWED_METADATA_TYPES])}"

    response = adaboard.request_adaboard(
        path="registry/metadata", params={"get_archived": False}
    ).json()
    if include_archived:
        res_archived = adaboard.request_adaboard(
            path="registry/metadata", params={"get_archived": True}
        ).json()
        response.extend(res_archived)

    type_repos = [x["name"] for x in response if x["type"] == type_id]

    return list(set(type_repos))


def get_repository_tags(project_name: str, repository_name: str) -> list[str]:
    """
    Get all tags of an image in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_repository_tags.ipynb) to test this function or build upon it.

    :param project_name: the name of the project the image belongs to
    :type project_name: str
    :param repository_name: the name of the image repository
    :type: repository_name: str
    :return: list of existing image tags
    :rtype: list
    """

    # Make sure repository options are valid
    assert (
        project_name in ALLOWED_OCI_PROJECT_TYPES
    ), f"Can only list tags for images that are of the type {', '.join([x for x in ALLOWED_OCI_PROJECT_TYPES])}"

    # Get all images from the repository, page by page
    all_images = adaboard.get_all_pages(
        api_url=f"registry/{project_name}/{repository_name}/images/v2/"
    )

    return [response["tags"][0] for response in all_images]


def get_types() -> list[str]:
    """
    Get a list of the image types in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/get_types.ipynb) to test this function or build upon it.

    :return: list of existing types
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path="registry/metadata",
    ).json()

    types = [x["type"] for x in response]

    return list(set(types))


def request_harbor(
    path: str,
    method=requests.get,
    **kwargs,
) -> requests.models.Response:
    """
    Function to request the registry API at the specified URL.

    Raises on errors from the HTTP calls.

    Note: All kwargs passed are directly sent through to the request function.

    :param path: path to query the registry on (base url will be injected
                 automatically)
    :type path: str
    :param method: method to use to query harbor, defaults to requests.get
    :type method: function, optional
    :return: response from harbor
    :rtype: response
    """
    # Construct authentication header for harbor
    adalib_config = config.get_config()
    harbor_token = keycloak.get_client_token(
        audience_client_id=adalib_config.KEYCLOAK_CLIENTS["harbor"]
    )

    headers = {
        "authorization": f"Bearer {harbor_token['id_token']}",
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    # Query harbor and raise for status
    constructed_url = (
        f"{adalib_config.SERVICES['harbor']['url']}/api/v2.0/{path}"
    )
    response = method(constructed_url, headers=headers, **kwargs)
    response.raise_for_status()

    return response


def restore_image(metadata_id: int) -> None:
    """
    Restore a previously-archived image in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/restore_image.ipynb) to test this function or build upon it.

    :param metadata_id: the metadata ID of the image
    :type metadata_id: int
    :return: nothing
    :rtype: None
    """

    # Check if image is already visible
    img_metadata = adaboard.request_adaboard(
        path=f"registry/metadata/{metadata_id}",
    ).json()

    # If not visible, restore it
    if img_metadata["archived"]:
        payload = {"archived": False}
        adaboard.request_adaboard(
            path=f"registry/metadata/{metadata_id}/archive",
            method=requests.put,
            json=payload,
        )
        return None
    else:
        raise RuntimeError(
            f"Image with metadata ID {metadata_id} is already visible."
        )


def update_image_state(metadata_id: str) -> str:
    """
    Update the "ready" state of an image based on its status in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/harbor/update_image_state.ipynb) to test this function or build upon it.

    :param metadata_id: the metadata ID of the image
    :type metadata_id: int
    :return: the updated "ready" state of the image
    :rtype: str
    """

    response = adaboard.request_adaboard(
        path=f"registry/metadata/{metadata_id}/ready", method=requests.put
    ).json()

    return response["ready"]



"""The Keywords sub-package exposes the core integrations of the keywords in the AdaLab Gallery.
"""

from .keywords import (
    create_keywords,
    delete_keywords,
    delete_keywords_by_id,
    get_keyword_id,
    get_keywords,
    get_keywords_stats,
    merge_keywords,
    rename_keyword,
)

__all__ = [
    "create_keywords",
    "delete_keywords",
    "delete_keywords_by_id",
    "get_keyword_id",
    "get_keywords",
    "get_keywords_stats",
    "merge_keywords",
    "rename_keyword",
]

__title__ = "adalib Keywords"



"""The Apps sub-package exposes the core integrations of the deployed apps in AdaLab."""

from .apps import (
    delete_app,
    deploy_app,
    deploy_multicontainer_app,
    edit_app,
    edit_multicontainer_app,
    get_all_apps,
    get_app,
    get_app_id,
    get_app_logs,
    get_apps_by_author,
    get_apps_status,
    get_container_idx,
    restart_app,
    start_app,
    stop_app,
)

__all__ = [
    "delete_app",
    "deploy_app",
    "deploy_multicontainer_app",
    "edit_app",
    "edit_multicontainer_app",
    "get_all_apps",
    "get_app",
    "get_app_id",
    "get_app_logs",
    "get_apps_by_author",
    "get_apps_status",
    "get_container_idx",
    "restart_app",
    "start_app",
    "stop_app",
]
__title__ = "adalib Apps"



import requests

from .. import adaboard


def create_keywords(
    new_keywords: str | list[str],
) -> list[dict[str, str | int]]:
    """
    Add a new keyword to the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/admin/keywords/create_keywords.ipynb) to test this function or build upon it.

    :param new_keyword: name (or names) of the keywords to be added
    :type new_keyword: str or list
    :return: a list of dictionaries with the new keyword ids
    :rtype: list
    """

    if not isinstance(new_keywords, list):
        new_keywords = [new_keywords]

    new_keyword_ids = []
    for new_keyword in new_keywords:
        payload = {"keyword": new_keyword}
        response = adaboard.request_adaboard(
            path="keywords",
            method=requests.post,
            json=payload,
        ).json()

        new_keyword_ids.append(
            {"keyword": new_keyword, "keyword_id": response["id"]}
        )
    return new_keyword_ids


def delete_keywords(keywords: str | list[str]) -> None:
    """
    Delete keywords from the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/admin/keywords/delete_keywords.ipynb) to test this function or build upon it.

    :param keywords: keywords to be deleted (tip: all names must be valid, otherwise the process fails)
    :type keywords: str or list
    :return: Nothing
    :rtype: None
    """

    if not isinstance(keywords, list):
        keywords = [keywords]

    keyword_ids = []
    for keyword in keywords:
        kw_id = get_keyword_id(keyword=keyword)
        keyword_ids.append(str(kw_id))
    delete_keywords_by_id(keyword_ids=keyword_ids)

    return None


def delete_keywords_by_id(keyword_ids: str | list[str]) -> None:
    """
    Delete keywords from the Gallery based on their ids.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/admin/keywords/delete_keywords_by_id.ipynb) to test this function or build upon it.

    :param keyword_ids: ids of the keywords to be deleted (tip: all ids must be valid, otherwise the process fails)
    :type keyword_ids: str or list
    :return: Nothing
    :rtype: None
    """

    if not isinstance(keyword_ids, list):
        keyword_ids = [keyword_ids]

    payload = {"keyword_ids": keyword_ids}
    adaboard.request_adaboard(
        path="keywords",
        method=requests.delete,
        json=payload,
    )

    return None


def get_keyword_id(keyword: str) -> int | None:
    """
    Find the ID for a specific keyword.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/keywords/get_keyword_id.ipynb) to test this function or build upon it.

    :param keyword: name of the keyword
    :type keyword: str
    :return: id of the keyword, or None if keyword not found
    :rtype: int or None
    """

    kws = get_keywords()
    for kw in kws:
        if keyword == kw["keyword"]:
            return kw["keyword_id"]
    return None


def get_keywords() -> list[dict[str, str | int]]:
    """
    Get information about all keywords in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/keywords/get_keywords.ipynb) to test this function or build upon it.

    :return: list of dictionaries with keywords data
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path="keywords",
        method=requests.get,
    ).json()

    return response


def get_keywords_stats() -> list[dict[str, str | int]]:
    """
    Get utilization statistics for the keywords in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/keywords/get_keywords_stats.ipynb) to test this function or build upon it.

    :return: list of dictionaries with keywords stats
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path="keywords/statistics",
        method=requests.get,
    ).json()

    return response


def merge_keywords(keywords: list[str], new_keyword: str) -> int:
    """
    Merge several keywords into a single one.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/admin/keywords/merge_keywords.ipynb) to test this function or build upon it.

    :param keywords: keywords to be merged
    :type keywords: list
    :param new_keyword: name of the new keyword (note: it can be one of the keywords to be merged)
    :type new_keyword: str
    :return: the new keyword's ID
    :rtype: int
    """

    keyword_ids = []
    for keyword in keywords:
        kw_id = get_keyword_id(keyword=keyword)
        keyword_ids.append(kw_id)

    payload = {"keyword_ids": keyword_ids, "new_keyword": new_keyword}
    response = adaboard.request_adaboard(
        path="keywords/merge",
        method=requests.put,
        json=payload,
    ).json()

    return response["id"]


def rename_keyword(keyword: str, new_keyword: str) -> int | str:
    """
    Rename a keyword.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/admin/keywords/rename_keyword.ipynb) to test this function or build upon it.

    :param keyword: the keyword to be renamed
    :type keyword: str
    :param new_keyword: name of the new keyword
    :type new_keyword: str
    :return: new keyword's id
    :rtype: int
    """

    keyword_id = get_keyword_id(keyword=keyword)
    payload = {"keyword_ids": [keyword_id], "new_keyword": new_keyword}
    response = adaboard.request_adaboard(
        path="keywords/merge",
        method=requests.put,
        json=payload,
    ).json()

    return response["id"]



"""The Superset sub-package exposes the core integrations with the Superset data platform connected to AdaLab."""

from .client import OAuthSupersetClient


def get_client(token=None):
    """Create a client for communicating with the Superset API

    :param token: An OAuth token valid for authenticating with the Superset API.
                  Optional if we're running in JupyterHub
    :return: authenticated Superset client
    :rtype: OAuthSupersetClient
    """
    return OAuthSupersetClient(oauth_token=token)


_superset_client = get_client()

databases = _superset_client.databases
datasets = _superset_client.datasets
saved_queries = _superset_client.saved_queries
charts = _superset_client.charts
dashboards = _superset_client.dashboards



"""Utilities for ACL validation and handling."""

from typing import Literal


def validate_acl(
    obj: Literal["app", "card", "schedule", "volume"],
    allow_none: bool = False,
    *acls: list[str],
):
    """Validate ACL objects, based on the given object type.

    :param obj: The object type to validate the ACLs for
    :type obj: Literal["app", "card", "schedule"]
    :param allow_none: Allow None as a valid ACL, defaults to False
    :type allow_none: bool
    :raises AssertionError: If the ACLs are invalid
    """
    allowed_acls = {
        "app": ["public", "logged_in", "userlist", "grouplist"],
        "card": ["public", "logged_in", "userlist", "grouplist"],
        "schedule": ["public", "logged_in", "userlist", "grouplist"],
        "volume": ["public", "logged_in", "userlist", "grouplist"],
    }

    assert obj in allowed_acls, f"Object does not support ACL: {obj}"
    for _acl in acls:
        if allow_none and _acl is None:
            continue

        assert (
            _acl.lower() in allowed_acls[obj]
        ), f"Invalid ACL value for {obj}: '{_acl}'. Allowed values: {allowed_acls[obj]}"



import requests

from .. import adaboard


def get_picture(picture_id: int, output_file_path: str) -> None:
    """
    Get a picture from the picture database.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/pictures/get_picture.ipynb) to test this function or build upon it.

    :param picture_id: ID of the picture
    :type picture_id: int
    :param output_file_path: path to save the picture
    :type output_file_path: str
    :return: None
    :rtype: None
    """

    response = adaboard.request_adaboard(
        path=f"images/{picture_id}",
        method=requests.get,
    )

    content_type = response.headers.get("Content-Type")
    # Ensure that the content type matches the extension of the output file
    if content_type == "image/jpeg":
        assert output_file_path.endswith(
            ".jpg"
        ), f"Output file path must end with '.jpg' for content type {content_type}"
    elif content_type == "image/png":
        assert output_file_path.endswith(
            ".png"
        ), f"Output file path must end with '.png' for content type {content_type}"
    elif content_type == "image/svg+xml":
        assert output_file_path.endswith(
            ".svg"
        ), f"Output file path must end with '.svg' for content type {content_type}"

    with open(output_file_path, "wb") as f:
        f.write(response.content)

    return None


def post_picture(input_file_path: str) -> int:
    """
    Post an picture to the picture database.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/pictures/post_picture.ipynb) to test this function or build upon it.

    :param input_file_path: path to the picture file
    :type input_file_path: str
    :return: ID of the posted picture
    :rtype: int
    """

    filename = input_file_path.split("/")[-1]
    content_type = infer_type_from_extension(filename)
    with open(input_file_path, "rb") as image_file:
        files = {"image_file": (filename, image_file, content_type)}

        response = adaboard.request_adaboard(
            path="image",
            method=requests.post,
            omit_content_type=True,
            files=files,
        ).json()

    return response["id"]


def post_picture_url(picture_url: str) -> int:
    """
    Post a picture to the picture database using a URL.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/picutres/post_picture_url.ipynb) to test this function or build upon it.

    :param picture_url: URL of the picture to be uploaded
    :type picture_url: str
    :return: ID of the posted picture
    :rtype: int
    """

    # Prepare the payload for the POST request
    payload = {"image_url": picture_url}

    # Send the request to the FastAPI endpoint
    response = adaboard.request_adaboard(
        path="image_url",  # The endpoint defined in your FastAPI application
        method=requests.post,
        json=payload,  # Use json parameter to send the payload as JSON
    ).json()

    # Return the ID of the newly created image
    return response["id"]


def infer_type_from_extension(filename: str) -> str:
    if filename.endswith(".jpg"):
        return "image/jpeg"
    elif filename.endswith(".png"):
        return "image/png"
    elif filename.endswith(".svg"):
        return "image/svg+xml"
    else:
        raise ValueError(
            f"Unsupported file extension for file {filename}. Supported extensions are '.jpg', '.png', and '.svg'."
        )



"""The Issues sub-package exposes the core integrations of the issues in the AdaLab Gallery.
"""

from .issues import create_issue, get_issue_data, get_issues, update_issue

__all__ = [
    "create_issue",
    "get_issue_data",
    "get_issues",
    "update_issue",
]

__title__ = "adalib Issues"



from datetime import datetime

import requests

from .. import adaboard
from ..utils import validate_acl, validate_volume_mount

PRIVILEGED_ROLES = ["Curator", "PlatformAdmin"]
ALLOWED_CONTAINER_KEYS = [
    "command",
    "environment_variables",
    "max_cpu",
    "max_ram",
    "metadata_id",
    "min_cpu",
    "min_ram",
    "port",
    "volume_mounts",
    "is_primary",
]
ALLOWED_LOG_SOURCES = ["container", "system"]
LOG_SOURCES = {"container": False, "system": True}


def delete_app(app_id: str) -> None:
    """
    Delete a deployed app from AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/delete_app.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    """

    adaboard.request_adaboard(path=f"apps/{app_id}/", method=requests.delete)
    return None


def deploy_app(
    name: str,
    description: str,
    metadata_id: int,
    url_path_prefix: str,
    stripped_prefix: bool = True,
    port: int = 80,
    environment_variables: dict = {},
    max_cpu: float = 1.0,
    min_cpu: float = 0.0,
    max_ram: int = 500,
    min_ram: int = 20,
    command: str = "",
    acl_type: str = "public",
    acl_list: list[str] = [],
    idp_enabled: bool = False,
    idp_scope: str = "",
    maintainers: list[str] = [],
    volume_mounts: list[dict[str,]] = [],
) -> str:
    """
    Deploy a single-container app to AdaLab from an existing metadata object.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/deploy_app.ipynb) to test this function or build upon it.

    :param name: the app's name
    :type name: str
    :param description: the app's description
    :type description: str
    :param metadata_id: the metadata ID of the container image
    :type metadata_id: int
    :param url_path_prefix: URL endpoint to deploy the app to
    :type url_path_prefix: str
    :param stripped_prefix: whether to strip the apps/{app_url} prefix from the URL, defaults to True
    :type stripped_prefix: bool, optional
    :param port: port to expose the app on, defaults to 80
    :type port: int, optional
    :param environment_variables: environment variables for the container, defaults to {}. Each entry must follow the format {"key": value}
    :type environment_variables: dict, optional
    :param max_cpu: maximum CPU usage allowed (vCPU), defaults to 1.0
    :type max_cpu: float, optional
    :param min_cpu: minimum CPU usage allowed (vCPU), defaults to 0.0
    :type min_cpu: float, optional
    :param max_ram: maximum RAM usage allowed (Mb), defaults to 500
    :type max_ram: int, optional
    :param min_ram: minimum RAM usage allowed (Mb), defaults to 20
    :type min_ram: int, optional
    :param command: command to start up the container, defaults to ""
    :type command: str, optional
    :param acl_type: type of access control list, defaults to "public"
    :type acl_type: str, optional
    :param acl_list: list of users or groups allowed to access the app when acl_type="userlist" or grouplist
    :type acl_list: list, optional
    :param idp_enabled: whether to enable IdP token in app headers, defaults to False
    :type idp_enabled: bool, optional
    :param idp_scope: IdP token scope, defaults to ""
    :type idp_scope: str, optional
    :param maintainers: list of users allowed to edit the app, defaults to []
    :type maintainers: list, optional
    :param volume_mounts: list of volume mounts for the container, defaults to []
    :type volume_mounts: list, optional
    :return: the app's ID
    :rtype: str
    """

    # Check that the app configuration is valid
    validate_acl("app", False, acl_type)
    for vm in volume_mounts:
        validate_volume_mount(volume_mount=vm)

    # Build the payload for the request
    payload = {
        "name": name,
        "description": description,
        "containers": [
            {
                "command": command,
                "environment_variables": [
                    {"key": k, "value": v}
                    for k, v in environment_variables.items()
                ],
                "max_cpu": max_cpu,
                "max_ram": max_ram,
                "metadata_id": metadata_id,
                "min_cpu": min_cpu,
                "min_ram": min_ram,
                "port": port,
                "is_primary": True,
                "volume_mounts": volume_mounts,
            }
        ],
        "url_path_prefix": url_path_prefix,
        "acls": [
            {
                "acl_action": "app_open",
                "acl_type": acl_type or "public",
                "userlist": acl_list if acl_type == "userlist" else [],
                "grouplist": acl_list if acl_type == "grouplist" else [],
            }
        ],
        "idp_enabled": idp_enabled,
        "idp_scope": idp_scope,
        "maintainers": maintainers,
        "stripped_prefix": stripped_prefix,
    }

    # Deploy the app
    response = adaboard.request_adaboard(
        path="apps/deploy/", method=requests.post, json=payload
    ).json()

    # Return the ID of the deployed app
    return response["id"]


def deploy_multicontainer_app(
    name: str,
    description: str,
    containers: list[dict[str,]],
    url_path_prefix: str,
    acl_type: str = "public",
    acl_list: list[str] = [],
    idp_enabled: bool = False,
    idp_scope: str = "",
    maintainers: list[str] = [],
    stripped_prefix: bool = True,
) -> str:
    """
    Deploy a multi-container app to AdaLab from an existing metadata object.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/deploy_multicontainer_app.ipynb) to test this function or build upon it.

    :param name: the app's name
    :type name: str
    :param description: the app's description
    :type description: str
    :param containers: list of container configurations. Each container must have a "metadata_id" key and can have the following optional keys: "command", "environment_variables", "max_cpu", "max_ram", "min_cpu", "min_ram", "port". Each entry in "environment_variables" must follow the format {"key": value}
    :type containers: list
    :param url_path_prefix: URL endpoint to deploy the app to
    :type url_path_prefix: str
    :param acl_type: type of access control list, defaults to "public"
    :type acl_type: str, optional
    :param acl_list: list of users or groups allowed to access the app when acl_type="userlist" or "grouplist", defaults to []
    :type acl_list: list, optional
    :param idp_enabled: whether to enable IdP token in app headers, defaults to False
    :type idp_enabled: bool, optional
    :param idp_scope: IdP token scope, defaults to ""
    :type idp_scope: str, optional
    :param maintainers: list of users allowed to edit the app, defaults to []
    :type maintainers: list, optional
    :param stripped_prefix: whether to strip the apps/{app_url} prefix from the URL, defaults to True
    :type stripped_prefix: bool, optional
    :return: the app's ID
    :rtype: str
    """

    # Check that the app configuration is valid
    validate_acl("app", False, acl_type)

    found_primary = False
    for container in containers:
        assert (
            "metadata_id" in container
        ), "Metadata ID of the container is required."
        assert "is_primary" in container, "Primary container flag is required."
        if container["is_primary"]:
            assert not found_primary, "Only one primary container is allowed."
            found_primary = True

        for key in container.keys():
            assert (
                key in ALLOWED_CONTAINER_KEYS
            ), f"Invalid container parameter key: {key}. Allowed keys are {', '.join(ALLOWED_CONTAINER_KEYS)}."

        for vm in container.get("volume_mounts", []):
            validate_volume_mount(volume_mount=vm)

    # Build the payload for the request
    payload = {
        "name": name,
        "description": description,
        "url_path_prefix": url_path_prefix,
        "stripped_prefix": stripped_prefix,
        "acls": [
            {
                "acl_action": "app_open",
                "acl_type": acl_type or "public",
                "userlist": acl_list if acl_type == "userlist" else [],
                "grouplist": acl_list if acl_type == "grouplist" else [],
            }
        ],
        "idp_enabled": idp_enabled,
        "idp_scope": idp_scope,
        "maintainers": maintainers,
    }

    for container in containers:
        container["environment_variables"] = [
            {"key": k, "value": v}
            for k, v in container.get("environment_variables", {}).items()
        ]

    payload["containers"] = containers

    # Deploy the app
    response = adaboard.request_adaboard(
        path="apps/deploy/", method=requests.post, json=payload
    ).json()

    # Return the ID of the deployed app
    return response["id"]


def edit_app(
    app_id: str,
    name: str = None,
    description: str = None,
    metadata_id: int = None,
    url_path_prefix: str = None,
    stripped_prefix: bool = None,
    port: int = None,
    environment_variables: dict = None,
    max_cpu: float = None,
    min_cpu: float = None,
    max_ram: int = None,
    min_ram: int = None,
    command: str = None,
    acl_type: str = None,
    acl_list: list[str] = None,
    idp_enabled: bool = None,
    idp_scope: str = None,
    maintainers: list[str] = None,
    volume_mounts: list[dict[str,]] = None,
) -> None:
    """
    Edit a single-container app's configuration.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/edit_app.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    :param name: the app's name, defaults to None (old one)
    :type name: str, optional
    :param description: the app's description, defaults to None (old one)
    :type description: str, optional
    :param metadata_id: the ID of the metadata object, defaults to None (old one)
    :type metadata_id: int, optional
    :param url_path_prefix: URL endpoint to deploy the app to, defaults to None (old one)
    :type url_path_prefix: str, optional
    :param stripped_prefix: whether to strip the apps/{app_url} prefix from the URL, defaults to None (old one)
    :type stripped_prefix: bool, optional
    :param port: the port where the app is served from, defaults to None (old one)
    :type port: int, optional
    :param replicas: number of replicas to deploy, defaults to None (old one)
    :type replicas: int, optional
    :param environment_variables: environment variables to pass to the app, defaults to None (old one)
    :type environment_variables: dict, optional
    :param max_cpu: maximum CPU usage allowed (vCPU), defaults to None (old one)
    :type max_cpu: float, optional
    :param min_cpu: minimum CPU usage allowed (vCPU), defaults to None (old one)
    :type min_cpu: float, optional
    :param max_ram: maximum RAM usage allowed (Mb), defaults to None (old one)
    :type max_ram: int, optional
    :param min_ram: minimum RAM usage allowed (Mb), defaults to None (old one)
    :type min_ram: int, optional
    :param command: command to start up the app, defaults to None (old one)
    :type command: str, optional
    :param acl_type: type of access control list, defaults to None (old one)
    :type acl_type: str, optional
    :param acl_list: list of users or groups allowed to access the app when acl_type="userlist" or "grouplist", defaults to []
    :type acl_list: list, optional
    :param idp_enabled: whether to enable IdP token in app headers, defaults to None (old one)
    :type idp_enabled: bool, optional
    :param idp_scope: IdP token scope, defaults to None (old one)
    :type idp_scope: str, optional
    :param maintainers: list of users allowed to edit the app, defaults to None (old one)
    :type maintainers: list, optional
    :param volume_mounts: list of volume mounts for the container, defaults to []
    :type volume_mounts: list, optional
    :return: nothing
    :rtype: None
    """

    # Collect configuration options
    environment_variables = environment_variables or dict()
    validate_acl("app", True, acl_type)

    if volume_mounts is not None:
        for vm in volume_mounts:
            validate_volume_mount(volume_mount=vm)

    if acl_type in ["userlist", "grouplist"]:
        assert (
            acl_list is not None
        ), "ACL type is userlist or grouplist but acl_list was not specified."

    acls = None
    if acl_type is not None:
        acls = [
            {
                "acl_action": "app_open",
                "acl_type": acl_type,
                "userlist": acl_list if acl_type == "userlist" else [],
                "grouplist": acl_list if acl_type == "grouplist" else [],
            }
        ]

    app_config = {
        "name": name,
        "description": description,
        "containers": [
            {
                "command": command,
                "environment_variables": (
                    [
                        {"key": k, "value": v}
                        for k, v in environment_variables.items()
                    ]
                    or None
                ),
                "max_cpu": max_cpu,
                "max_ram": max_ram,
                "metadata_id": metadata_id,
                "min_cpu": min_cpu,
                "min_ram": min_ram,
                "port": port,
                "is_primary": True,
                "volume_mounts": volume_mounts,
            }
        ],
        "url_path_prefix": url_path_prefix,
        "acls": acls,
        "idp_enabled": idp_enabled,
        "idp_scope": idp_scope,
        "maintainers": maintainers,
        "stripped_prefix": stripped_prefix,
    }

    # Fetch old configuration
    old_config = adaboard.request_adaboard(
        path=f"apps/{app_id}/info/", method=requests.get
    ).json()

    # Build request payload combining new and old options
    containers = [
        {
            k: v if v is not None else old_config["containers"][0][k]
            for k, v in app_config["containers"][0].items()
        }
    ]
    payload = {
        k: v if v is not None else old_config[k] for k, v in app_config.items()
    }
    payload["containers"] = containers

    adaboard.request_adaboard(
        path=f"apps/{app_id}/update/", method=requests.put, json=payload
    )

    return None


def edit_multicontainer_app(
    app_id: str,
    name: str = None,
    description: str = None,
    containers: list[dict[str,]] = None,
    url_path_prefix: str = None,
    acl_type: str = None,
    acl_list: list[str] = None,
    idp_enabled: bool = None,
    idp_scope: str = None,
    stripped_prefix: bool = None,
) -> None:
    """
    Edit a multi-container app's configuration.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/edit_multicontainer_app.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    :param name: the app's name, defaults to None (old one)
    :type name: str, optional
    :param description: the app's description, defaults to None (old one)
    :type description: str, optional
    :param containers: list of container configurations, defaults to None (old one)
    :type containers: list, optional
    :param url_path_prefix: URL endpoint to deploy the app to, defaults to None (old one)
    :type url_path_prefix: str, optional
    :param acl_type: type of access control list, defaults to None (old one)
    :type acl_type: str, optional
    :param acl_list: list of users or groups allowed to access the app when acl_type="userlist" or "grouplist", defaults to None (old one)
    :type acl_list: list, optional
    :param idp_enabled: whether to enable IdP token in app headers, defaults to None (old one)
    :type idp_enabled: bool, optional
    :param idp_scope: IdP token scope, defaults to None (old one)
    :type idp_scope: str, optional
    :param stripped_prefix: whether to strip the apps/{app_url} prefix from the URL, defaults to None (old one)
    :type stripped_prefix: bool, optional
    :return: nothing
    :rtype: None
    """
    validate_acl("app", True, acl_type)
    if acl_type in ["userlist", "grouplist"]:
        assert (
            acl_list is not None
        ), "ACL type is userlist or grouplist but acl_list was not specified."

    acls = None
    if acl_type is not None:
        acls = [
            {
                "acl_action": "app_open",
                "acl_type": acl_type,
                "userlist": acl_list if acl_type == "userlist" else [],
                "grouplist": acl_list if acl_type == "grouplist" else [],
            }
        ]

    # Collect configuration options
    app_config = {
        "name": name,
        "description": description,
        "containers": containers,
        "url_path_prefix": url_path_prefix,
        "acls": acls,
        "idp_enabled": idp_enabled,
        "idp_scope": idp_scope,
        "stripped_prefix": stripped_prefix,
    }

    # Check that the container configurations are valid
    if containers:
        found_primary = False
        for container in containers:
            assert (
                "metadata_id" in container
            ), "Metadata ID of the container is required."
            assert (
                "is_primary" in container
            ), "Primary container flag is required."
            if container["is_primary"]:
                assert (
                    not found_primary
                ), "Only one primary container is allowed."
                found_primary = True

            for key in container.keys():
                assert (
                    key in ALLOWED_CONTAINER_KEYS
                ), f"Invalid container parameter key: {key}. Allowed keys are {', '.join(ALLOWED_CONTAINER_KEYS)}."

            for vm in container.get("volume_mounts", []):
                validate_volume_mount(volume_mount=vm)

            container["environment_variables"] = [
                {"key": k, "value": v}
                for k, v in container.get(
                    "environment_variables", dict()
                ).items()
            ]

    # Fetch old configuration
    old_config = adaboard.request_adaboard(
        path=f"apps/{app_id}/info/", method=requests.get
    ).json()
    payload = {
        k: v if v is not None else old_config[k] for k, v in app_config.items()
    }
    adaboard.request_adaboard(
        path=f"apps/{app_id}/update/", method=requests.put, json=payload
    )

    return None


def get_all_apps() -> (
    list[dict[str, str | int | bool | dict[str, str] | list[str]]]
):
    """
    Get information for all the apps deployed in AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/get_all_apps.ipynb) to test this function or build upon it.

    :return: list with all the information of each app
    :rtype: list
    """

    # If the user is privileged enough, get all apps
    response_roles = adaboard.request_adaboard(path="users/self").json()
    include_all = any(
        role in PRIVILEGED_ROLES
        for role in response_roles["roles"]["adaboard"]
    )
    # Get the list of all apps
    response = adaboard.request_adaboard(
        path="apps/all/", method=requests.get, params={"all": include_all}
    ).json()

    return response


def get_app(
    app_id: str,
) -> dict[str, str | int | bool | dict[str, str] | list[str]]:
    """
    Get the information of a specific deployed app in AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/get_app.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    :return: app's information
    :rtype: dict
    """

    response = adaboard.request_adaboard(
        path=f"apps/{app_id}/info/", method=requests.get
    ).json()

    return response


def get_app_id(
    app_name: str = "", author_id: str = "", app_url: str = ""
) -> str:
    """
    Get the ID of a deployed app in AdaLab. Either app_url, or app_name and author_id, must be specified.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/get_app_id.ipynb) to test this function or build upon it.

    :param app_name: name of the app, defaults to ""
    :type app_name: str, optional
    :param author_id: author of the app, defaults to ""
    :type author_id: str, optional
    :param app_url: endpoint of the app's URL, defaults to ""
    :type app_url: str, optional
    :return: the app's ID
    :rtype: str
    """

    # Check that enough information is provided to find the app
    assert (
        app_name and author_id
    ) or app_url, (
        "Either app_url, or app_name and author_id, must be specified."
    )
    # If the user is privileged enough, get all apps
    response_roles = adaboard.request_adaboard(path="users/self").json()
    include_all = any(
        role in PRIVILEGED_ROLES
        for role in response_roles["roles"]["adaboard"]
    )
    # Get the list of all apps
    response = adaboard.request_adaboard(
        path="apps/all/", method=requests.get, params={"all": include_all}
    ).json()

    # Return the ID of the requested app
    try:
        if app_url:
            return [
                app["app_id"]
                for app in response
                if app["url_path_prefix"] == app_url
            ][0]
        else:
            return [
                app["app_id"]
                for app in response
                if app["name"] == app_name and app["user_id"] == author_id
            ][0]
    except IndexError:
        raise ValueError("No app found with the specified parameters.")


def get_app_logs(
    app_id: str,
    container_image_name: str = "",
    container_metadata_id: int = 0,
    container_primary: bool = True,
    container_idx: int = 0,
    from_date: str = "",
    to_date: str = "",
    source: str = "container",
) -> str | list[dict[str, str]]:
    """
    Get the container or system logs of a deployed app in AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/get_app_logs.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    :param container_image_name: name of the container image to get logs from
    :type container_image_name: str, optional
    :param container_metadata_id: the metadata ID of the container image to get logs from, defaults to 0
    :type container_metadata_id: int, optional
    :param container_primary: whether to get logs from the primary container, defaults to True
    :type container_primary: bool, optional
    :param container_idx: index of the container to get logs from, defaults to 0
    :type container_idx: int, optional
    :param from_date: start date for the logs, defaults to ""
    :type from_date: str, optional
    :param to_date: end date for the logs, defaults to ""
    :type to_date: str, optional
    :param source: source of the logs, either "container" or "system", defaults to "container"
    :type source: str, optional
    :return: deployment logs
    :rtype: str
    """

    assert (
        source in ALLOWED_LOG_SOURCES
    ), f"Log source must be one of {', '.join([x for x in ALLOWED_LOG_SOURCES])}."
    assert not (
        container_image_name and container_metadata_id and container_idx
    ), "Either container image name, container metadata ID or index must be specified."
    assert not (
        container_image_name and container_metadata_id
    ), "Only one container image identifier allowed."

    # If container name provided, get the index in the app
    if container_image_name and not container_primary:
        tmp = get_app(app_id)
        app_mdata_id = [
            container["metadata_id"] for container in tmp["containers"]
        ]
        imgs = adaboard.request_adaboard(
            path="registry/metadata", params={"get_archived": False}
        ).json()
        img_names = [
            {"name": img["name"], "metadata_id": img["metadata_id"]}
            for img in imgs
            if img["name"] == container_image_name
        ]
        cnt_mdata_id = [
            img["metadata_id"]
            for img in img_names
            if img["metadata_id"] in app_mdata_id
        ]
        assert not (len(cnt_mdata_id) == 0), "Container not found in app."
        assert not (
            len(cnt_mdata_id) > 1
        ), "Multiple container images found with the same name."
        container_idx = [
            idx
            for idx, cnt in enumerate(tmp["containers"])
            if cnt["metadata_id"] in cnt_mdata_id
        ][0]

    if container_metadata_id and not container_primary:
        tmp = get_app(app_id)
        app_mdata_id = [
            container["metadata_id"] for container in tmp["containers"]
        ]
        cnt_mdata_id = [
            idx
            for idx, x in enumerate(app_mdata_id)
            if x == container_metadata_id
        ]
        assert not (len(cnt_mdata_id) == 0), "Container not found in app."
        assert not (
            len(cnt_mdata_id) > 1
        ), "Multiple containers found with the same container image metadata ID."
        container_idx = cnt_mdata_id[0]

    if container_primary:
        tmp = get_app(app_id=app_id)
        for idx, container in enumerate(tmp["containers"]):
            if container["is_primary"]:
                container_idx = idx
                break

    params = {"container_idx": container_idx, "system": LOG_SOURCES[source]}

    # Check that the date format is correct and update the request
    if from_date:
        try:
            datetime.fromisoformat(from_date)
        except ValueError:
            raise AssertionError(
                "Date must be in ISO format (YYYY-MM-DDTHH:mm:ss+HH:mm)."
            )
        params["start_dt"] = from_date
    if to_date:
        try:
            datetime.fromisoformat(to_date)
        except ValueError:
            raise AssertionError(
                "Date must be in ISO format (YYYY-MM-DDTHH:mm:ss+HH:mm)."
            )
        params["end_dt"] = to_date

    response = adaboard.request_adaboard(
        path=f"apps/v2/{app_id}/logs", method=requests.get, params=params
    ).json()

    return response


def get_apps_by_author(
    author_id: str,
) -> list[dict[str, str | int | bool | dict[str, str] | list[str]]]:
    """
    Get information for all the apps deployed in AdaLab by a specific author.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/get_apps_by_author.ipynb) to test this function or build upon it.

    :param author_id: the ID of the app's author
    :type author_id: str
    :return: list with all the information of each app
    :rtype: list
    """

    # If the user is privileged enough, get all apps
    response_roles = adaboard.request_adaboard(path="users/self").json()
    include_all = any(
        role in PRIVILEGED_ROLES
        for role in response_roles["roles"]["adaboard"]
    )
    # Get the list of all apps
    response = adaboard.request_adaboard(
        path="apps/all/", method=requests.get, params={"all": include_all}
    ).json()

    return [app for app in response if app["user_id"] == author_id]


def get_apps_status() -> dict[str, str]:
    """
    Get the status of all the apps deployed in AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/get_apps_status.ipynb) to test this function or build upon it.

    :return: dictionary with the status of each app
    :rtype: dict
    """

    # Get status report for all apps
    response_stats = adaboard.request_adaboard(
        path="apps/status/", method=requests.get
    ).json()

    # If the user is privileged enough, get all apps
    response_roles = adaboard.request_adaboard(path="users/self").json()
    include_all = any(
        role in PRIVILEGED_ROLES
        for role in response_roles["roles"]["adaboard"]
    )
    # Get the list of all apps
    response_apps = adaboard.request_adaboard(
        path="apps/all/", method=requests.get, params={"all": include_all}
    ).json()

    # Build dictionary with the status of each app
    apps_status = [
        {
            "name": app["name"],
            "author": app["user_id"],
            "URL": app["url_path_prefix"],
            "status": response_stats[str(app["status"])],
        }
        for app in response_apps
    ]
    return apps_status


def get_container_idx(app_id: str, metadata_id: int) -> int:
    """Get the index of a specific container of a deployed app in AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/get_container_idx.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    :param metadata_id: the metadata ID of the container
    :type metadata_id: int
    :return: the index of the container in the app
    :rtype: int
    """

    response = adaboard.request_adaboard(
        path=f"apps/{app_id}/info/", method=requests.get
    ).json()

    for idx, container in enumerate(response["containers"]):
        if container["metadata_id"] == metadata_id:
            return idx

    assert False, "No container found with the specified metadata ID."


def restart_app(app_id: str) -> None:
    """
    Restart a deployed app in AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/restart_app.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    :return: nothing
    :rtype: None
    """

    adaboard.request_adaboard(
        path=f"apps/{app_id}/restart/", method=requests.put
    )
    return None


def start_app(app_id: str) -> None:
    """
    Start a deployed app in AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/start_app.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    :return: nothing
    :rtype: None
    """

    adaboard.request_adaboard(
        path=f"apps/{app_id}/start/", method=requests.put
    )

    return None


def stop_app(app_id: str) -> None:
    """
    Stop a deployed app in AdaLab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/apps/stop_app.ipynb) to test this function or build upon it.

    :param app_id: the app's ID
    :type app_id: str
    :return: nothing
    :rtype: None
    """

    adaboard.request_adaboard(path=f"apps/{app_id}/stop/", method=requests.put)
    return None



import datetime as dt

import requests

from .. import adaboard


def create_issue(card_id: int, message: str, username: str) -> int:
    """
    Create a new issue for a specific card in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/issues/create_issue.ipynb) to test this function or build upon it.

    :param card_id: id of the card to assign the issue to
    :type card_id: int
    :param message: message of the new issue
    :type message: str
    :param username: the current username (tip: check "LOGNAME" environment variable)
    :type username: str
    :return: id of the new issue
    :rtype: int
    """

    timestamp = dt.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    payload = {
        "card_id": card_id,
        "message": message,
        "reported_by": username,
        "reported_date": timestamp,
        "status": "new",
    }
    response = adaboard.request_adaboard(
        path="cardissues",
        method=requests.post,
        json=payload,
    ).json()

    return response["id"]


def get_issue_data(
    issue_id: int,
) -> dict[
    str,
    int
    | str
    | dict[str, str | int]
    | list[dict[str, int | str | dict[str, int | str]]],
]:
    """
    Get contents and information of a specific issue.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/issues/get_issue_data.ipynb) to test this function or build upon it.

    :param issue_id: id of the issue whose info is to be fetched
    :type issue_id: int
    :return: dictionary with the issue's detailed info
    :rtype: dict
    """

    response = adaboard.request_adaboard(
        path=f"cardissues/{issue_id}", method=requests.get
    ).json()

    return response


def get_issues() -> list[dict[str, int | str | dict[str, str | int]]]:
    """
    Get all the card issues registered in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/issues/get_issues.ipynb) to test this function or build upon it.

    :return: list of dictionaries with issues' main info
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path="cardissues", method=requests.get
    ).json()

    return response


def update_issue(
    issue_id: int, new_status: str = "", comment: str = ""
) -> int:
    """
    Update issue content (status, comments).
    Note that only one of the fields can be updated at a time.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/issues/update_issue.ipynb) to test this function or build upon it.

    :param issue_id: the id of the issue to be updated
    :type issue_id: int
    :param new_status: status to be set
    :type new_status: str
    :param comment: comment to be added
    :type comment: str
    :return: id of the issue update
    :rtype: int
    """

    if comment and not new_status:
        payload = {
            "comment": comment,
            "history_type": "comment",
        }
    elif new_status and not comment:
        assert new_status.lower() in [
            "new",
            "in progress",
            "in_progress",
            "solved",
            "ignored",
        ], "Invalid status!"
        payload = {
            "new_status": new_status.lower().replace(" ", "_"),
            "comment": "",
            "history_type": "new_status",
        }
    else:
        return "Only one field can be updated at a time!"

    response = adaboard.request_adaboard(
        path=f"cardissues/{issue_id}",
        method=requests.put,
        json=payload,
    ).json()

    return response["id"]



import base64
from datetime import datetime
from typing import Optional

import requests
from adalib_auth import config

from .. import adaboard

ALLOWED_GIT_TYPES = ["ssh", "https", "public"]
ALLOWED_BUILD_LOG_SOURCES = ["build", "system"]
ALLOWED_LAB_LOG_SOURCES = ["user", "system"]
ALLOWED_OCI_PROJECT_TYPES = ["apps", "kernels", "base_images"]
HARBOR_HOST = "container-registry"
HARBOR_PORT = "80"
LOG_SOURCES = {"user": False, "system": True}


def build_image_from_git(
    git_url: str,
    file_path: str,
    image_name: str,
    image_tag: str,
    image_type: str,
    branch: str = "",
    build_args: dict[str, str] = {},
    commit: str = "",
    git_creds: dict[str, str] = {},
    project_name: str = "",
    timeout: int = 3600,
) -> int:
    """
    Build a container image from a container file in a Git repository.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/build_image_from_git.ipynb) to test this function or build upon it.

    :param git_url: full URL of Git repository containing the container file
    :type git_url: str
    :param file_path: path to the container file in the Git repository
    :type file_path: str
    :param image_name: the name of the target image
    :type image_name: str
    :param image_tag: the tag of the target image
    :type image_tag: str
    :param image_type: the type of the image
    :type image_type: str
    :param branch: the branch of the Git repository to use, defaults to default
    :type branch: str, optional
    :param build_args: additional build arguments to pass build process, defaults to {}
    :type build_args: dict, optional
    :param commit: the commit hash of the Git repository to use, defaults to ""
    :type commit: str, optional
    :param git_creds: credentials to authenticate to the Git repository. For SSH, use keys "PRIVATE_KEY_FILE" and "PASSWORD". For HTTPS, use keys "USERNAME" and "PASSWORD".
    :type git_creds: dict, optional
    :param project_name: the name of the project to link the image to, defaults to image_name
    :type project_name: str, optional
    :param timeout: the maximum time (in seconds) to wait for the build to complete, defaults to 3600
    :type timeout: int, optional
    :return: the run ID of the build process
    :rtype: int
    """

    adalib_config = config.get_config()
    # Check that the specified project is valid
    assert (
        image_type in ALLOWED_OCI_PROJECT_TYPES
    ), f"Can only build images that are of the type {', '.join([x for x in ALLOWED_OCI_PROJECT_TYPES])}"
    # Check that only branch or commit is specified
    assert not (
        bool(branch) and bool(commit)
    ), "Cannot specify both 'branch' and 'commit'"
    # Check that the file path is valid
    file_name = file_path.split("/")[-1]
    assert ("dockerfile" in file_name.lower()) or (
        "containerfile" in file_name.lower()
    ), "The file path must point at a 'Dockerfile' or 'Containerfile' type of file."
    # Figure out git type
    if git_creds:
        if list(git_creds.keys()) == ["PRIVATE_KEY_FILE", "PASSWORD"]:
            git_type = "ssh"
        elif list(git_creds.keys()) == ["USERNAME", "PASSWORD"]:
            git_type = "https"
        else:
            raise KeyError(
                "Dictionary with git credentials has the wrong keys. For SSH, use 'PRIVATE_KEY_FILE' and 'PASSWORD'. For HTTPS, use 'USERNAME' and 'PASSWORD'."
            )
    else:
        git_type = "public"

    # Build request payload with the specified parameters
    # Prepare build arguments
    build_args = ";;".join([f"{k}={v}" for k, v in build_args.items()])

    # Collect parameters into payload
    payload = {
        "script": "builder",
        "user_id": adaboard.get_user()["user_id"],
        "pool": "internal",
        "cleanup": True,
        "timeout": timeout,
        "run_in_isolation": False,
        "start_podman_sidecar": False,
        "config": {
            "HARBOR_CURATED_PROJECT": {
                "real_value": image_type,
                "value": image_type,
            },
            "HARBOR_HOST": {
                "real_value": HARBOR_HOST,
                "value": HARBOR_HOST,
            },
            "HARBOR_HUMAN_READABLE_REPOSITORY_NAME": {
                "real_value": project_name or image_name,
                "value": project_name or image_name,
            },
            "HARBOR_NAMESPACE": {
                "real_value": adalib_config.NAMESPACE,
                "value": adalib_config.NAMESPACE,
            },
            "HARBOR_PROJECT": {
                "real_value": f"{image_type}_temp",
                "value": f"{image_type}_temp",
            },
            "HARBOR_REPOSITORY_NAME": {
                "real_value": image_name,
                "value": image_name,
            },
            "HARBOR_TAG": {
                "real_value": image_tag,
                "value": image_tag,
            },
            "IMAGE_BUILD_ARGS": {
                "real_value": build_args,
                "value": build_args,
            },
            "IMAGE_CONTAINERFILE_LOCATION": {
                "real_value": file_path,
                "value": file_path,
            },
            "NETWORK_HOST": {
                "real_value": adalib_config.NETWORK_HOST,
                "value": adalib_config.NETWORK_HOST,
            },
        },
    }

    # Add Git credentials to payload based on Git type
    if git_type.lower() == "ssh":
        git_url = git_url.replace("https://", "git@").replace(".com/", ".com:")
        git_config = [
            {
                "GIT_URL": {"real_value": git_url, "value": git_url},
                "GIT_SSH_PRIVATE_KEY_FILE": {
                    "real_value": git_creds["PRIVATE_KEY_FILE"],
                    "value": git_creds["PRIVATE_KEY_FILE"],
                },
                "GIT_SSH_PASSWORD": {
                    "real_value": git_creds["PASSWORD"],
                    "value": git_creds["PASSWORD"],
                },
            },
        ]
    elif git_type.lower() == "https":
        git_config = [
            {
                "GIT_URL": {"real_value": git_url, "value": git_url},
                "GIT_USERNAME": {
                    "real_value": git_creds["USERNAME"],
                    "value": git_creds["USERNAME"],
                },
                "GIT_PASSWORD": {
                    "real_value": git_creds["PASSWORD"],
                    "value": git_creds["PASSWORD"],
                },
            },
        ]
    elif git_type.lower() == "public":
        git_config = [
            {
                "GIT_URL": {"real_value": git_url, "value": git_url},
            },
        ]

    # Use specific branch or commit
    if branch:
        git_config[0]["GIT_BRANCH"] = {"real_value": branch, "value": branch}
    elif commit:
        git_config[0]["GIT_COMMIT"] = {"real_value": commit, "value": commit}

    # Add Git configuration to payload
    for item in git_config:
        payload["config"].update(item)

    # Trigger the build and get the run ID
    response_build = adaboard.request_adaboard(
        path="script_runner/runs/start/",
        method=requests.post,
        json=payload,
    ).json()

    return response_build["id"]


def build_image_from_lab(
    file_path: str,
    image_name: str,
    image_tag: str,
    image_type: str,
    project_name: str = "",
    build_args: dict[str, str] = {},
    timeout: int = 3600,
) -> int:
    """
    Build a container image from a container file in the Lab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/build_image_from_lab.ipynb) to test this function or build upon it.

    :param file_path: path to the container file in the Lab
    :type file_path: str
    :param image_name: the name of the target image
    :type image_name: str
    :param image_tag: the tag of the target image
    :type image_tag: str
    :param image_type: the type of the image
    :type image_type: str
    :param project_name: the name of the project to link the image to, defaults to image_name
    :type project_name: str, optional
    :param build_args: additional build arguments to pass build process, defaults to {}
    :type build_args: dict, optional
    :param timeout: the maximum time (in seconds) to wait for the build to complete, defaults to 3600
    :type timeout: int, optional
    :return: the run ID of the build process
    :rtype: int
    """

    adalib_config = config.get_config()
    # Check that the specified project is valid
    assert (
        image_type in ALLOWED_OCI_PROJECT_TYPES
    ), f"Can only build images that are of the type {', '.join([x for x in ALLOWED_OCI_PROJECT_TYPES])}"

    # Get the context and file name from the full path
    ctxt_path = file_path.split("/")[0]
    file_name = file_path.split("/")[-1]

    assert ("dockerfile" in file_name.lower()) or (
        "containerfile" in file_name.lower()
    ), "The file path must point at a 'Dockerfile' or 'Containerfile' type of file."

    # Build request payload with the specified parameters
    build_args = ";;".join([f"{k}={v}" for k, v in build_args.items()])

    payload = {
        "script": "builder",
        "user_id": adaboard.get_user()["user_id"],
        "pool": "internal",
        "cleanup": True,
        "timeout": timeout,
        "run_in_isolation": False,
        "start_podman_sidecar": False,
        "config": {
            "GIT_LOCATION": {
                "real_value": ctxt_path,
                "value": ctxt_path,
            },
            "HARBOR_CURATED_PROJECT": {
                "real_value": image_type,
                "value": image_type,
            },
            "HARBOR_HOST": {
                "real_value": HARBOR_HOST,
                "value": HARBOR_HOST,
            },
            "HARBOR_HUMAN_READABLE_REPOSITORY_NAME": {
                "real_value": project_name or image_name,
                "value": project_name or image_name,
            },
            "HARBOR_NAMESPACE": {
                "real_value": adalib_config.NAMESPACE,
                "value": adalib_config.NAMESPACE,
            },
            "HARBOR_PROJECT": {
                "real_value": f"{image_type}_temp",
                "value": f"{image_type}_temp",
            },
            "HARBOR_REPOSITORY_NAME": {
                "real_value": image_name,
                "value": image_name,
            },
            "HARBOR_TAG": {
                "real_value": image_tag,
                "value": image_tag,
            },
            "IMAGE_BUILD_ARGS": {
                "real_value": build_args,
                "value": build_args,
            },
            "IMAGE_CONTAINERFILE_LOCATION": {
                "real_value": file_name,
                "value": file_name,
            },
            "NETWORK_HOST": {
                "real_value": adalib_config.NETWORK_HOST,
                "value": adalib_config.NETWORK_HOST,
            },
        },
    }

    # Trigger the build and get the run ID
    response_build = adaboard.request_adaboard(
        path="script_runner/runs/start/",
        method=requests.post,
        json=payload,
    ).json()

    return response_build["id"]


def delete_files(paths: list[str]) -> None:
    """
    Delete files or folders in the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/delete_files.ipynb) to test this function or build upon it.

    :param paths: paths to the files to delete, relative to the user's home directory
    :type paths: list
    :return: nothing
    :rtype: None
    """

    # Build request payload with the specified information
    payload = {"operation": "delete", "paths": paths}

    # Delete the files
    adaboard.request_adaboard(
        path="jupyter/files/notebook/content/",
        method=requests.patch,
        json=payload,
    )

    return None


def download_file(file_path: str, target_path: str) -> None:
    """
    Download a file from the user's Lab environment and save it to a specific location.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/download_file.ipynb) to test this function or build upon it.

    :param file_path: path to the file in the Lab environment, including file name and extension, relative to the user's home directory (/home/<username>)
    :type file_path: str
    :param target_path: path to save the file to, including file name and extension, relative to the current working directory
    :type target_path: str
    :return: nothing
    :rtype: None
    """

    # Get the file, base64 encoded
    response = adaboard.request_adaboard(
        path="jupyter/files/notebook/content/fetch", params={"path": file_path}
    ).json()

    # Decode the file to get the binary
    file_bytes = base64.b64decode(response["content"])

    # Write the file to the target path
    with open(target_path, "wb") as out_file:
        out_file.write(file_bytes)

    return None


def get_available_kernels() -> (
    list[dict[str, str | int | bool | dict[str, str]]]
):
    """
    Get a list of available kernel images in the Harbor registry.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_available_kernels.ipynb) to test this function or build upon it.

    :return: a list of available kernels
    :rtype: list
    """

    # Get metadata for all images
    response = adaboard.request_adaboard(path="registry/metadata").json()

    # Filter-out non-kernel images
    return [x for x in response if x["type"] == "kernel"]


def get_build_logs(
    build_id: int, source: str = "build"
) -> list[dict[str, str]]:
    """
    Get the build or system logs of an image build process.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_build_logs.ipynb) to test this function or build upon it.

    :param build_id: the run ID of the build process
    :type build_id: int
    :param source: the source of the logs, either "build" or "system", defaults to "build"
    :type source: str, optional
    :return: build logs
    :rtype: list
    """

    assert (
        source in ALLOWED_BUILD_LOG_SOURCES
    ), f"Log source must be one of {', '.join([x for x in ALLOWED_BUILD_LOG_SOURCES])}"

    response_raw = adaboard.request_adaboard(
        path=f"script_runner/runs/{build_id}/logs/"
    ).json()
    response = response_raw["log"]

    if source == "build":
        logs = [
            {
                "time": res["timestamp"],
                "input": res["input"],
                "output": res["output"].replace(res["input"], ""),
            }
            for res in response
            if res["type"] == "command"
        ]
    elif source == "system":
        logs = [
            {"time": res["timestamp"], "message": res["output"]}
            for res in response
            if res["type"] == "system"
        ]

    return logs


def get_build_status(
    build_id: int,
) -> str:
    """
    Get the status of a build process.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_build_status.ipynb) to test this function or build upon it.

    :param build_id: the run ID of the build process
    :type build_id: int
    :return: the status of the build process
    :rtype: str
    """

    response_statuses = adaboard.request_adaboard(path="runner/status/").json()

    response_run = adaboard.request_adaboard(
        path=f"script_runner/runs/{build_id}/"
    ).json()

    return response_statuses[str(response_run["status"])]


def get_config_options() -> (
    list[dict[str, int | str | list[dict[str, str | int | bool]]]]
):
    """
    Get the available Lab configuration options for the user.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_config_options.ipynb) to test this function or build upon it.

    :return: list of existing Lab configuration options
    :rtype: list
    """

    response = adaboard.request_adaboard(path="jupyter/system/options").json()

    return response


def get_installed_kernels() -> list[dict[str, str | int]]:
    """
    Get a list of the kernels that are installed in the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_installed_kernels.ipynb) to test this function or build upon it.

    :return: a list of installed kernels
    :rtype: list
    """

    response = adaboard.request_adaboard(path="jupyter/kernelspecs").json()

    return response


def get_kernel_metadata_id(kernel_name: str, kernel_tag: str) -> int:
    """
    Get the metadata ID of the object corresponding to a specific kernel image.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_kernel_metadata_id.ipynb) to test this function or build upon it.

    :param kernel_name: the name of the repository containing the kernel image
    :type kernel_name: str
    :param kernel_tag: the tag of the kernel image
    :type kernel_tag: str
    :return: the metadata ID of the kernel image
    :rtype: int
    """

    # Get metadata for all images
    response = adaboard.request_adaboard(path="registry/metadata").json()

    # Filter-out non-kernel images
    kernel_images = [x for x in response if x["type"] == "kernel"]

    # Return the metadata ID of the specified kernel image, if it exists
    try:
        return [
            x["metadata_id"]
            for x in kernel_images
            if x["oci_image_name"]["repository"] == kernel_name
            and x["oci_image_name"]["tag"] == kernel_tag
        ][0]
    except IndexError:
        raise ValueError(
            f"Kernel image {kernel_name}:{kernel_tag} does not exist."
        )


def get_lab_files(path: str) -> list[dict[str, str | int]]:
    """
    Get a list of the files under a directory in the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_lab_files.ipynb) to test this function or build upon it.

    :param path: path to list the files from, relative to the user's home directory (/home/<username>)
    :type path: str
    :return: files in the directory
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path="jupyter/files/notebook/content/", params={"path": path}
    ).json()

    return response


def get_lab_images() -> list[dict[str, str]]:
    """
    Get a list of container images present in the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_lab_images.ipynb) to test this function or build upon it.

    :return: a list of available images
    :rtype: list
    """

    response = adaboard.request_adaboard(path="registry/jupyter_images").json()

    return response


def get_lab_logs(
    from_date: str = "", to_date: str = "", source: str = "user"
) -> list[dict[str, str]]:
    """
    Get the logs of the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_lab_logs.ipynb) to test this function or build upon it.

    :param from_date: the start date for the logs, in ISO format (YYYY-MM-DDTHH:mm:ss+HH:mm), defaults to ""
    :type from_date: str, optional
    :param to_date: the end date for the logs, in ISO format (YYYY-MM-DDTHH:mm:ss+HH:mm), defaults to ""
    :type to_date: str, optional
    :param source: the source of the logs, either "user" or "system", defaults to "user"
    :type source: str, optional
    :return: Lab logs
    :rtype: list
    """

    # Check that the specified source is valid
    assert (
        source in ALLOWED_LAB_LOG_SOURCES
    ), f"Log source must be one of {', '.join([x for x in ALLOWED_LAB_LOG_SOURCES])}"

    params = {"system": LOG_SOURCES[source]}
    # Check that the date format is correct
    if from_date:
        try:
            datetime.fromisoformat(from_date)
        except ValueError:
            raise AssertionError(
                "Date must be in ISO format (YYYY-MM-DDTHH:mm:ss+HH:mm)."
            )
        params["start_dt"] = from_date

    if to_date:
        try:
            datetime.fromisoformat(to_date)
        except ValueError:
            raise AssertionError(
                "Date must be in ISO format (YYYY-MM-DDTHH:mm:ss+HH:mm)."
            )
        params["end_dt"] = to_date

    response = adaboard.request_adaboard(
        path="jupyter/server/v2/logs", params=params
    ).json()

    return response


def get_lab_status(metrics: bool = True, options: bool = True) -> dict[
    str,
    str
    | dict[str, float]
    | list[dict[str, str | int | float | dict[str, str | int | bool]]],
]:
    """
    Get configuration and status information about the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/get_lab_status.ipynb) to test this function or build upon it.

    :param metrics: whether to fetch Lab metrics, defaults to True
    :type metrics: bool, optional
    :param options: whether to fetch Lab options, defaults to True
    :type options: bool, optional
    :return: Lab information
    :rtype: dict
    """

    response = adaboard.request_adaboard(
        path="jupyter/server",
        params={"metrics": metrics, "options": options},
    ).json()

    return response


def install_kernel(
    metadata_id: int,
    name: str,
    include_notebook: bool = True,
) -> None:
    """
    Install a kernel into the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/install_kernel.ipynb) to test this function or build upon it.

    :param metadata_id: the metadata ID of the kernel to install
    :type metadata_id: int
    :param name: name to be given to the installed kernel
    :type name: str
    :param include_notebook: whether to include a dummy notebook file with the kernel, defaults to True
    :type include_notebook: bool, optional
    :return: nothing
    :rtype: None
    """

    # Build request payload with the specified information
    payload = {
        "display_name": name + " (published)",
        "metadata_id": metadata_id,
        "include_dummy_notebook": include_notebook,
    }

    # Install kernel and check response
    adaboard.request_adaboard(
        path="jupyter/kernelspecs", method=requests.post, json=payload
    ).json()

    return None


def move_file(old_path: str, new_path: str) -> None:
    """
    Move a file or folder in the user's Lab environment.  Note that this operation can also be used to rename a file or folder.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/move_file.ipynb) to test this function or build upon it.

    :param old_path: path to the file to move, including file name and extension, relative to the user's home directory
    :type old_path: str
    :param new_path: new path to save the file to, including file name and extension, relative to the user's home directory
    :type new_path: str
    :return: nothing
    :rtype: None
    """

    # Build request payload with the specified information
    payload = {
        "operation": "rename",
        "old_path": old_path,
        "new_path": new_path,
    }

    # Rename the file
    adaboard.request_adaboard(
        path="jupyter/files/notebook/content/",
        method=requests.patch,
        json=payload,
    )

    return None


def stop_lab() -> None:
    """
    Stop the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/stop_lab.ipynb) to test this function or build upon it.

    :return: nothing
    :rtype: None
    """

    response = adaboard.request_adaboard(
        path="jupyter/server", method=requests.delete
    ).json()

    assert (
        response["message"] == "OK"
    ), "The Lab environment could not be stopped."

    return None


def uninstall_kernel(name: str) -> None:
    """
    Uninstall a kernel from the user's Lab environment.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/uninstall_kernel.ipynb) to test this function or build upon it.

    :param name: the name of the kernel to uninstall
    :type name: str
    :return: nothing
    :rtype: None
    """

    response = adaboard.request_adaboard(
        path=f"jupyter/kernelspecs/{name}", method=requests.delete
    ).json()

    assert response["message"] == "OK", "The kernel could not be uninstalled."

    return None


def who_am_i() -> dict[str, str]:
    """
    Get information about the current user.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/lab/who_am_i.ipynb) to test this function or build upon it.

    :return: information about the current user
    :rtype: dict
    """

    response = adaboard.request_adaboard(path="users/self").json()

    return response


def convert_token_to_value(token: str) -> Optional[str]:
    """
    Convert a Lab Configuration token into the true decrypted value from the user input.

    :param token: the token to convert
    :type token: str
    :return: the true decrypted value of the token
    :rtype: str
    """
    response = adaboard.request_adaboard(
        path="jupyter/system/user_input_value", params={"token": token}
    ).json()
    return response.get("value")
